/***********************************************************************
 * Test bench for the ALU model.
 *
 **********************************************************************/

module alu_test
 import alu_opcodes::*;             // import ALU opcode definitions
(input  logic [3:0] result,
 output logic [3:0] a, b,
 output opcode_t    opcode,
 input  logic       test_clk
);
  timeunit 1ns; timeprecision 1ns;

  int error_count = 0;  // number of errors detected

  initial begin   // input stimulus
    a = 2; 
    b = 3;
    opcode = opcode.first;

    // STEP THROUGH ALL OPCODES
    @(posedge test_clk);
    for (int i = 0; i < opcode.num; i++) begin
      @(negedge test_clk) check_results();  // call output verification task
      opcode = opcode.next();
    end

    // An alternative way of doing the loop without a separate loop variable, using a do-while loop.
    //do 
    //begin
    //  @(negedge test_clk) check_results();  // call output verification task
    //  opcode = opcode.next();
    //end
    //while ( opcode != opcode.first());

    @(posedge test_clk);
    $display("\nALU TESTS COMPLETED WITH %0d ERRORS!\n", error_count);
    // $stop();
    $finish();
  end

  initial begin  // output verification
    //display time in ns, no decimal places, 7 character column width
    $timeformat(-9, 0, " ns", 7);
  end

  task check_results;
    logic [3:0] expected;    // local variable
    case (opcode)
      INVERT_A   :  expected = ~a;
      NEGATE_A   :  expected = -a;
      INCREMENT_A:  expected = a + 1;
      A_PLUS_B   :  expected = a + b;
      A_MINUS_B  :  expected = a - b;
      A_XORED_B  :  expected = a ^ b;
      A_SHIFTED_B:  expected = a >> (|b[3:2]? 4: b[1:0]);  // shift b times; max of 4
      A_ROTATED_B:  expected = {a,a} >> b[1:0];  // rotate (b modulus 4) number of times
    endcase
    $write("At %t:  a=%h  b=%h  opcode=%-11s :  result=%h (%b)",
             $realtime, a, b, opcode.name, result, result);
    if (result === expected)
      $display(" - OK");
    else begin
      $display(" - ERROR: expected %h (%b)", expected, expected);
      error_count++;
    end
  endtask

endmodule: alu_test
