/***********************************************************************
 * A SystemVerilog testbench for an instruction register.
 * The course labs will convert this to an object-oriented testbench
 * with constrained random test generation, functional coverage, and
 * a scoreboard for self-verification.
 *
 **********************************************************************/

module instr_register_test
  import instr_register_pkg::*;  // user-defined types are defined in instr_register_pkg.sv
  (tb_ifc.tb io);  // interface port

  timeunit 1ns; timeprecision 1ns;

  initial begin
    $display("\nReseting the instruction register...");
    io.write_pointer <= 5'h00;         // initialize write pointer
    io.read_pointer  <= 5'h1F;         // initialize read pointer
    io.load_en       <= 1'b0;          // initialize load control line
    io.reset_n       <= 1'b0;          // assert reset_n (active low) - should be non-blocking
    repeat (2) @(posedge io.clk) ;     // hold in reset for 2 clock cycles
    io.reset_n       <= 1'b1;          // deassert reset_n (active low)

    $display("\nWriting values to register stack...");
    @(posedge io.clk)
    #1 io.load_en = 1'b1;  // enable writing to register
    repeat (3) begin
      @(posedge io.clk) 
      #1 randomize_transaction();
      #8 print_transaction();
    end
    @(posedge io.clk) 
    #1 io.load_en = 1'b0;  // turn-off writing to register

    // read back and display same three register locations
    $display("\nReading back the same register locations written...");
    for (int i=0; i<3; i++) begin
      // Later labs will replace this loop with iterating through a
      // scoreboard to determine which addresses were written and
      // the expected values to be read back
      @(posedge io.clk);
      #1 io.read_pointer = i;
      #8 print_results();
    end

    @(posedge io.clk) $finish();
  end

  function void randomize_transaction();
    // A later lab will replace this function with SystemVerilog
    // constrained random values
    //
    // The static temp variable is required in order to write to fixed
    // addresses of 0, 1 and 2.  This will be replaced with randomized
    // write_pointer values in a later lab
    //
    static int temp = 0;
    io.operand_a     <= $urandom_range(0, 30) - 15;       // between -15 and 15
    io.operand_b     <= $urandom_range(0, 15);            // between 0 and 15
    io.opcode        <= opcode_t'($urandom_range(0, 7));  // between 0 and 7, cast to opcode_t type
    io.write_pointer <= temp++;
  endfunction: randomize_transaction
  
  function void print_transaction();
    $display("Write to register location %2d: opcode = %0d (%5s) operand_a =%3d  operand_b =%3d", 
 	     io.write_pointer, io.opcode, io.opcode.name(), io.operand_a, io.operand_b);
  endfunction

  function void print_results();
    $display("Read from register location %2d: opcode = %0d (%5s) operand_a =%3d  operand_b =%3d", 
	     io.read_pointer, io.instruction_word.opc, io.instruction_word.opc.name(), io.instruction_word.op_a, io.instruction_word.op_b);
  endfunction

endmodule: instr_register_test
