
class Scoreboard extends Component;

  mailbox #(Transaction) mon2scb_h;
  mailbox #(Transaction) gen2scb_h;
  
  Transaction sent_trn[int];
  Transaction rcvd_trn[int];
  Transaction trn_h;      // transaction handle for covergroup and transaction from generator
  Transaction mon_trn_h;  // transaction handle for transactdion from monitor

  int num_sent_trn, num_rcvd_trn;
  
  covergroup cg_transaction;
    // Cover write_pointer in four bins, where each bin covers one quadrant (1/4) of the address space
    ADDR: coverpoint trn_h.write_pointer {
      bins addr[4]={[0:$]};} // divide range into 4 bins

    // Cover opcode in two bins: DIV operations, and all other operations
    OPC: coverpoint trn_h.opcode {
      bins div_opc = { DIV };
      bins other_opc = default;
    }
    // Cover operand_a for values that are negative, positive, and zero
    OPA: coverpoint trn_h.operand_a {
      bins neg_A = {[-31:-1]};
      bins zero_A = { 0 };
      bins pos_A = {[1:31]};
    }
    // Cover operand_b for values that are negative, positive, and zero
    OPB: coverpoint trn_h.operand_b {
      bins neg_B = {[-31:-1]};
      bins zero_B = { 0 };
      bins pos_B = {[1:31]};
    }
    // Cover how many times a DIV operation occurs when operand_b is zero
    Div_zero: cross OPC, OPB {
      bins DIV_BY_ZERO = binsof(OPC.div_opc) && binsof(OPB.zero_B);
      ignore_bins rest = !binsof(OPC.div_opc) || !binsof(OPB.zero_B);
    }
  endgroup : cg_transaction

  function new();
    cg_transaction = new();                       // Construct cover group
  endfunction


/*
    // This function no longer needed.  The get can have the trn_h directly 
    // point to the transaction, which has been cloned at source, so ther eis no need 
    //to create a new one.  
    function void save_transaction(Transaction trn_h);
    // creates a shallow clone of trn object and stores handle to it in the array
    // the LRM shallow clone syntax is used, not a user defined clone method
    // In this syntax it is illegal to have parenthases after new.
    sent_trn[trn_h.write_pointer]=new trn_h; 
    this.trn_h = trn_h;
    cg_transaction.sample();                 // Sample coverage
  endfunction
*/

  task run(int n);
    // both mon2scb and gen2scb mailboxes are unbounded, so there is no issue with leaving 
    // transactions on the mailbox until ready to deal with them. 
    // not useing the write pointer as the index, but the transaction number, this overcomes
    // issues with multiple transactions with the same write pointer. 
    fork
      repeat (n) begin
        gen2scb_h.get(trn_h);
        sent_trn[num_sent_trn++] = trn_h;
        cg_transaction.sample();
      end
      repeat (n) begin
        mon2scb_h.get(mon_trn_h);
        rcvd_trn[num_rcvd_trn++] = mon_trn_h;
      end
    join
  endtask : run

  task check(int n);
    foreach(sent_trn[i]) begin
      if (!sent_trn[i].compare(rcvd_trn[i])) begin
        $error(" Expected Transaction %p \n           Actual Transaction %p", sent_trn[i], rcvd_trn[i]);
      end
    end
  endtask


  task report();
    $display("\n********Coverage report***********");
    $display("Checked %0d transactions", sent_trn.size);
    $display("Covergroup coverage     = %6.2f%%", cg_transaction.get_coverage());
    $display("Address coverage        = %6.2f%%", cg_transaction.ADDR.get_coverage());
    $display("Opcode coverage         = %6.2f%%", cg_transaction.OPC.get_coverage());
    $display("Operand a coverage      = %6.2f%%", cg_transaction.OPA.get_coverage());
    $display("Operand b coverage      = %6.2f%%", cg_transaction.OPB.get_coverage());
    $display("Divide-by-zero coverage = %6.2f%%", cg_transaction.Div_zero.get_coverage());
    $display("\nOverall coverage        = %6.2f%%", $get_coverage());
    $display("**********************************\n");
  endtask


endclass: Scoreboard